#!/usr/bin/env python
'''
This script allows you to easily generate a controller configuration file based
on the larpix chip key interface.
Generally, this script can generate string-based keys that contain the chip_id,
io channel, and io group within the key. The characters $, {, and } are reserved for special
functions and should not be used within keys. To use:
python gen_controller_config.py --name <config name> --key_fmt <formatting strings> --outfile <output file config file>

The formatting string is used to generate chip keys, generally it is a list of
strings that specify chip keys. Using ${} special functions allow you to
generate chip keys according to a integer list. For example:
 - --key_fmt 'test-${0-2}' generate 3 chip keys (test-0, test-1, test-2)
 - --key_fmt 'test-${0,1,2}' generates the same
 - --key_fmt 'test-${0-1,2}' also generates the same
For more complex chip keys that have repeated sets of identical chip ids, you
can use multiple instances of the ${} special function. This operates with the
same syntax. For example:
 - --key_fmt 'test-${0-2}-${0}' generates 5 chip keys (test-0-0, test-1-0,
   test-2-0)
 - --key_fmt 'test-${0-2}-${0,1}' generates 6 chip keys (test-0-0, test-0-1,
 test-1-0, test-1-1, test-2-0, test-2-1)
Finally, the --key_fmt argument can be passed multiple strings to generate a
variety of chip key formats:
 - --key_fmt 'key-0-${0-2}' 'key-1-${2-4}' 'diff-${0-2}' generates 9 chip keys
   (key-0-0, key-0-1, key-0-2, key-1-2, key-1-3, key-1-4, diff-0, diff-1,
   diff-2)

'''
import argparse
import os
import json
from itertools import zip_longest

from larpix import configs
import larpix

parser = argparse.ArgumentParser(usage=__doc__)
parser.add_argument('--name', '-n', type=str, required=True, help='''
    configuration name (required)
    ''')
parser.add_argument('--key_fmt', '-k', type=str, required=True, nargs='+',
    help='''
    key formatting specifier (required)
    see general help for more information
    ''')
parser.add_argument('--outfile', '-o', type=str, required=False, default=None,
    help='''
    output filename (optional, default=<name>_chip_info.json)
    if file exists, chip keys will be appended to the config file
    ''')
# parser.add_argument('--io', type=str, required=False, default=None, help='''
#     io class to validate generated keys against (optional, default=%(default)s)
#     path should be relative to the larpix.io module, e.g. 'zmq_io.ZMQ_IO' is a
#     valid path specifier
#     ''')
args = parser.parse_args()

# Load up arguments
name = args.name
key_fmts = args.key_fmt
# if not all([key_fmt.count('$[') == 1 for key_fmt in key_fmts]):
#     raise RuntimeError('A single chip key must be specified in each key formatting string')
outfile = args.outfile
if outfile is None:
    outfile = '{}_chip_info.json'.format(name)
io = None
if not io is None:
    module_str = args.io.split('.')
    io_module = None
    if len(module_str) > 1:
        io_module = __import__('larpix.io.'+'.'.join(module_str[:-1]), fromlist=[module_str[-1]])
    else:
        io_module = __import__('larpix.io', fromlist=[module_str[-1]])
    io = getattr(io_module, module_str[-1])

# Load existing configuration
configuration = {}
if os.path.exists(outfile):
    with open(outfile,'r') as ifile:
        configuration = json.load(ifile)
    if not io is None and 'chip_list' in configuration.keys():
        for chip_key, chip_id in configuration['chip_list']:
            if not io.is_valid_chip_key(chip_key):
                raise RuntimeError('Previous configuration file key \'{}\' is incompatible with io  {}'.format(chip_key, io))
else:
    print('No existing file {} found, generating from scratch...'.format(outfile))

# Declare helper functions
def append_chip_ids(sub_keys, chip_id_fmt):
    '''
    Appends all specified chip_ids to subkey strings
    Specification of chip_ids follows 2,4,5-7 = [2,4,5,6,7]
    Returns all new subkeys along with associated chip ids as a 2-tuple

    '''
    return_keys = []
    return_ids = []
    chip_id_sub_strs = chip_id_fmt.split(',')
    for chip_id_sub_str in chip_id_sub_strs:
        chip_id_spec = chip_id_sub_str.split('-')
        if len(chip_id_spec) == 2:
            chip_ids  = range(int(chip_id_spec[0]), int(chip_id_spec[1])+1)
            for chip_id in chip_ids:
                for sub_key in sub_keys:
                    return_keys += [sub_key + str(chip_id)]
                    return_ids += [chip_id]
        else:
            for sub_key in sub_keys:
                return_keys += [sub_key + str(int(chip_id_spec[0]))]
                return_ids += [int(chip_id_spec[0])]
        # print('end loop chip_id', chip_id_sub_str, return_keys, return_ids)
    return return_keys, return_ids

def append_int(sub_keys, chip_ids, int_fmt):
    '''
    Appends all specified integers to subkey strings
    Specification of integers follows 2,4,5-7 = [2,4,5,6,7]
    Returns all new subkeys along with associated chip ids as a 2-tuple

    '''
    return_keys = []
    return_ids = []
    int_sub_strs = int_fmt.split(',')
    for int_sub_str in int_sub_strs:
        int_spec = int_sub_str.split('-')
        if len(int_spec) == 2:
            ints = range(int(int_spec[0]), int(int_spec[1])+1)
            for integer in ints:
                # print(sub_keys)
                # print(chip_ids)
                for sub_key, chip_id in zip_longest(sub_keys, chip_ids):
                    return_keys += [sub_key + str(integer)]
                    # print(return_keys)
                    if chip_ids:
                        return_ids += [chip_id]
        else:
            for sub_key, chip_id in zip_longest(sub_keys, chip_ids):
                return_keys += [sub_key + str(int(int_spec[0]))]
                if chip_ids:
                    return_ids += [chip_id]
        # print('end loop int', int_sub_str, return_keys, return_ids)
    return return_keys, return_ids

def multi_split(strings, delimiters=[' ','\t','\n']):
    '''
    Recursively splits a list of strings according to a list of delimiters
    E.g. multi_split(['test, word1', ', word2'],delimiters=[',',' ']) returns
    ['test','word1','word2']

    '''
    if len(delimiters) == 1:
        return_strings = []
        for string in strings:
            return_strings += string.split(delimiters[0])
        return return_strings
    else:
        return_strings = strings
        for delimiter in delimiters:
            return_strings = multi_split(return_strings, delimiter)
        return return_strings

def generate_keys(key_fmt):
    '''
    Generates all possible chip keys and chip ids based on string formatter

    '''
    return_keys = ['']
    return_ids = []

    # print(key_fmt)
    # str_segments = multi_split([key_fmt], delimiters=['$',']','}'])
    str_segments = multi_split([key_fmt], delimiters=['$','}'])
    # print(str_segments)
    for str_seg in str_segments:
        if not str_seg:
            continue
        # if str_seg[0] == '[':
        #     return_keys, return_ids = append_chip_ids(return_keys, str_seg[1:])
        if str_seg[0] == '{':
            return_keys, return_ids = append_int(return_keys, return_ids, str_seg[1:])
        elif return_keys:
            for idx, key in enumerate(return_keys):
                new_key = key + str_seg
                return_keys[idx] = new_key
        # print('end loop', str_seg, return_keys, return_ids)
    return return_keys, return_ids

# Initialize
configuration['name'] = name
if not 'chip_list' in configuration.keys():
    configuration['chip_list'] = []
# Generate keys
for key_fmt in key_fmts:
    chip_keys, chip_ids = generate_keys(key_fmt)
    # for chip_key, chip_id in zip(chip_keys, chip_ids):
    for chip_key in chip_keys:
        # Check if key is valid for io class
        if not io is None:
            if not io.is_valid_chip_key(chip_key):
                raise RuntimeError('Format specifier led to an invalid chip key \'{}\''.format(chip_key))
        # Check if chip_key is unique (overwrite chip id, if exists)
        existing_keys = [chip_key for chip_key in configuration['chip_list']]
        if chip_key in existing_keys:
            idx = list(zip(*configuration['chip_list']))[0].index(chip_key)
            configuration['chip_list'][idx] = chip_key
        else:
            configuration['chip_list'].append(chip_key)

# Save
with open(outfile, 'w') as of:
    json.dump(configuration, of, indent=4)




from bitarray import bitarray
import struct

from .. import bitarrayhelper as bah
from ..key import Key
from . import Packet

class PacketCollection(object):
    '''
    Represents a group of packets that were sent to or received from
    LArPix.

    Index into the PacketCollection as if it were a list:

        >>> collection[0]
        Packet(b'\x07\x00\x00\x00\x00\x00\x00')
        >>> first_ten = collection[:10]
        >>> len(first_ten)
        10
        >>> type(first_ten)
        larpix.larpix.PacketCollection
        >>> first_ten.message
        'my packets | subset slice(None, 10, None)'

    To view the bits representation, add 'bits' to the index:

        >>> collection[0, 'bits']
        '00000000 00000000 00000000 00000000 00000000 00000000 000111'
        >>> bits_format_first_10 = collection[:10, 'bits']
        >>> type(bits_format_first_10[0])
        str


    '''
    def __init__(self, packets, bytestream=None, message='',
            read_id=None, skipped=None):
        self.packets = packets
        self.bytestream = bytestream
        self.skipped = skipped
        self.message = message
        self.read_id = read_id
        self.parent = None

    def __eq__(self, other):
        '''
        Return True if the packets, message and bytestream compare equal.

        '''
        return (self.packets == other.packets and
                self.message == other.message and
                self.bytestream == other.bytestream)

    def __repr__(self):
        return '<%s with %d packets, read_id %d, "%s">' % (self.__class__.__name__,
                len(self.packets), self.read_id, self.message)

    def __str__(self):
        if len(self.packets) < 20:
            return '\n'.join(str(packet) for packet in self.packets)
        else:
            beginning = '\n'.join(str(packet) for packet in self.packets[:10])
            middle = '\n'.join(['   .', '   . omitted %d packets' %
                (len(self.packets)-20), '   .'])
            end = '\n'.join(str(packet) for packet in self.packets[-10:])
            return '\n'.join([beginning, middle, end])

    def __len__(self):
        return len(self.packets)

    def __getitem__(self, key):
        '''
        Get the specified item(s).

        If key is an int, return the packet object at that index in
        self.packets.

        If key is a slice, return a PacketCollection with the specified
        packets, and with a message inherited from self.message.

        If key is (slice or int, 'str'), use the behavior as if setting
        key = key[0].

        If key is (int, 'bits'), return a string representation of the
        bits of the specified packet, as determined by
        self._bits_getitem.

        If key is (slice, 'bits'), return a list of string
        representations of the packets specified by the slice.

        '''
        if isinstance(key, slice):
            items = PacketCollection([p for p in self.packets[key]])
            items.message = '%s | subset %s' % (self.message, key)
            items.parent = self
            items.read_id = self.read_id
            return items
        elif isinstance(key, tuple):
            if key[1] == 'bits':
                return self._bits_getitem(key[0])
            elif key[1] == 'str':
                return self[key[0]]
        else:
            return self.packets[key]

    def _bits_getitem(self, key):
        '''
        Replace each packet with a string of the packet bits grouped 8
        bits at a time.

        '''
        if isinstance(key, slice):
            return [' '.join(p.bits.to01()[i:i+8] for i in
                range(0, p.size, 8)) for p in self.packets[key]]
        else:
            p = self.packets[key]
            return ' '.join(p.bits.to01()[i:i+8] for i in
                    range(0, p.size, 8))

    def to_dict(self):
        '''
        Export the information in this PacketCollection to a dict.

        '''
        d = {}
        d['packets'] = [packet.export() for packet in self.packets]
        d['id'] = id(self)
        d['parent'] = 'None' if self.parent is None else id(self.parent)
        d['message'] = str(self.message)
        d['read_id'] = 'None' if self.read_id is None else self.read_id
        d['bytestream'] = ('None' if self.bytestream is None else
                self.bytestream.decode('raw_unicode_escape'))
        return d

    def from_dict(self, d):
        '''
        Load the information in the dict into this PacketCollection.

        '''
        self.message = d['message']
        self.read_id = d['read_id']
        self.bytestream = d['bytestream'].encode('raw_unicode_escape')
        self.parent = None
        self.packets = []
        for p in d['packets']:
            bits = p['bits']
            packet = Packet()
            packet.bits = bitarray(bits)
            self.packets.append(packet)

    def extract(self, *attrs, **selection):
        '''
        Extract the given attribute(s) from packets specified by selection
        and return a list.

        Any attribute of a Packet is a valid attribute or selection

        Usage:

        >>> # Return a list of adc counts from any data packets
        >>> dataword = collection.extract('dataword', packet_type=0)
        >>> # Return a list of timestamps from chip 2 data
        >>> timestamps = collection.extract('timestamp', chip_id=2, packet_type=Packet_v2.DATA_PACKET)
        >>> # Return the most recently read global threshold from chip 5
        >>> threshold = collection.extract('register_value', register_address=32, packet_type=3, chip_id=5)[-1]
        >>> # Return multiple attributes
        >>> chip_keys, channel_ids = zip(*collection.extract('chip_key','channel_id'))

        .. note:: selecting on ``timestamp`` will also select
            TimestampPacket values.
        '''
        values = []
        for p in self.packets:
            try:
                if all(getattr(p,key) == value for key, value in selection.items()):
                    if len(attrs) > 1:
                        values.append([getattr(p,attr) for attr in attrs])
                    else:
                        values.append(getattr(p,attrs[0]))
            except AttributeError:
                continue
        return values

    def origin(self):
        '''
        Return the original PacketCollection that this PacketCollection
        derives from.

        '''
        child = self
        parent = self.parent
        max_generations = 100  # to prevent infinite loops
        i = 0
        while parent is not None and i < max_generations:
            # Move up the family tree one generation
            child = parent
            parent = parent.parent
            i += 1
        if parent is None:
            return child
        else:
            raise ValueError('Reached limit on generations: %d' %
                    max_generations)

    def with_chip_key(self, chip_key):
        '''
        Return packets with the specified chip key.

        '''
        return [packet for packet in self.packets \
            if packet.chip_key == chip_key]

    def by_chip_key(self):
        '''
        Return a dict of { chip_key: PacketCollection }.

        '''
        chip_groups = {}
        for packet in self.packets:
            # append packet to list if list exists, else append to empty
            # list as a default
            key = packet.chip_key
            chip_groups.setdefault(key, []).append(packet)
        to_return = {}
        for chip_key in chip_groups:
            new_collection = PacketCollection(chip_groups[chip_key])
            new_collection.message = self.message + ' | chip {}'.format(chip_key)
            new_collection.read_id = self.read_id
            new_collection.parent = self
            to_return[chip_key] = new_collection
        return to_return

    def with_chipid(self, chipid):
        '''
        Return packets with the specified chip ID.

        '''
        return [packet for packet in self.packets if packet.chipid == chipid]

    def by_chipid(self):
        '''
        Return a dict of { chipid: PacketCollection }.

        '''
        chip_groups = {}
        for packet in self.packets:
            # append packet to list if list exists, else append to empty
            # list as a default
            chip_groups.setdefault(packet.chip_id, []).append(packet)
        to_return = {}
        for chipid in chip_groups:
            new_collection = PacketCollection(chip_groups[chipid])
            new_collection.message = self.message + ' | chip %s' % chipid
            new_collection.read_id = self.read_id
            new_collection.parent = self
            to_return[chipid] = new_collection
        return to_return



import argparse
import time

import larpix
import larpix.format.rawhdf5format
import larpix.format.pacman_msg_format
import larpix.format.hdf5format
from larpix.format.rawhdf5format import from_rawfile, len_rawfile
from larpix.format.pacman_msg_format import parse
from larpix.format.hdf5format import to_file

def main(input_filename, output_filename, block_size):
    total_messages = len_rawfile(input_filename)
    total_blocks = total_messages // block_size + 1
    last = time.time()
    for i_block in range(total_blocks):
        start = i_block * block_size
        end = min(start + block_size, total_messages)
        if start == end: return

        if time.time() > last + 1:
            print('reading block {} of {}...\r'.format(i_block+1,total_blocks),end='')
            last = time.time()
        rd = from_rawfile(input_filename, start=start, end=end)
        pkts = list()
        for i_msg,data in enumerate(zip(rd['msg_headers']['io_groups'], rd['msgs'])):
            io_group,msg = data
            pkts.extend(parse(msg, io_group=io_group))
        to_file(output_filename, packet_list=pkts)
    print()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--input_filename', '-i', type=str, help='''Input hdf5 file, formatted with larpix.format.rawhdf5format using the larpix.io.PACMAN_IO class''')
    parser.add_argument('--output_filename', '-o', type=str, help='''Output hdf5 file,
        to be formatted with larpix.format.hdf5format''')
    parser.add_argument('--block_size', default=10240, type=int, help='''Max number of messages to store in working memory (default=%(default)s)''')
    args = parser.parse_args()
    c = main(**vars(args))



from bitarray import bitarray
import os
import errno
import functools
from collections import OrderedDict

from .. import bitarrayhelper as bah
from .. import configs
from . import BaseConfiguration, _Smart_List

class BaseConfiguration_v2(BaseConfiguration):
    '''
    Base class for building desired configurations of a LArPix v2 (or v2b) chip.

    Specific configurations (v2, v2b, lightpix-v1) inherit from this class, and their specific attributes can be added.

    '''
    _endian = 'little'
    num_channels = 64

    def __init__(self):
        super(BaseConfiguration_v2, self).__init__()
        return

    def all_data(self, endian='little'):
        bits = []
        for register_name in self.register_names:
            register_data =  getattr(self, register_name+'_data')
            for register_addr, register_bits in register_data:
                if len(bits) == register_addr:
                    if endian[0] == 'l':
                        bits += [register_bits]
                    else:
                        bits += [register_bits[::-1]]
        return bits

    def some_data(self, registers, endian='little'):
        '''
        Fetch register addresses and data from a selected set of registers

        :param registers: list of registers to fetch data for, specified either by register name (str) or register addresses (int)

        :returns: tuple of list of register addresses and list of register data

        '''
        bits = []
        addrs = []
        for register in registers:
            if isinstance(register, int):
                register_name = self.register_map_inv[register][0]
                register_data = getattr(self, register_name+'_data')
            elif isinstance(register, str):
                register_data = getattr(self, register+'_data')
            for register_addr, register_bits in register_data:
                if isinstance(register, int) and register_addr != register:
                    continue
                if endian[0] == 'l':
                    bits += [register_bits]
                    addrs += [register_addr]
                else:
                    bits += [register_bits[::-1]]
                    addrs += [register_addr]
        return addrs, bits

    def from_dict_registers(self, d, endian='little'):
        '''
        Load in the configuration specified by a dict of (register,
        value) pairs.

        '''
        for address, value in d.items():
            register_names = self.register_map_inv[address]
            for register_name in register_names:
                setattr(self, register_names[0] + '_data', (address, bah.fromuint(value,8,endian=endian)))
        return

    def _is_register_value_pair(self, item):
        '''
        Helper function to determine if item is a register, bitarray pair

        '''
        if isinstance(item, tuple) and len(item) == 2 and \
                isinstance(item[0], int) and isinstance(item[1], bitarray)\
                and len(item[1]) == 8 and item[0] >= 0 and item[0] < self.num_registers:
            return True
        return False

### Stuff for generating the v2 configuration properties

## Getter function formulas
#
def _basic_getter(register_name):
    '''
    Function formula for getting a named register

    '''
    def basic_getter_func(self):
        return getattr(self, '_'+register_name)
    return basic_getter_func
# /Getter function formulas

## Setter function formulas
#
def _basic_setter(register_name):
    '''
    Function formula for setting a named register

    '''
    def basic_setter_func(self, value):
        setattr(self, '_'+register_name, value)
    return basic_setter_func

def _list_setter(register_name, min_value, max_value):
    '''
    Function formula for setting a named register

    '''
    def list_setter_func(self, value):
        setattr(self, '_'+register_name, _Smart_List(value, min_value, max_value))
    return list_setter_func

# /Setter function formulas

## Data getter function formulas
#
def _basic_data_getter(register_name):
    '''
    Function formula for getting a "simple" register's data
    A simple register has a one-to-many relationship between the register
    name and register addresses.

    Returns a list of register_addr, bitarray pairs

    '''
    def basic_data_getter_func(self):
        register_range = self.register_map[register_name]
        value = getattr(self, register_name)
        all_data = bah.fromuint(value, 8*(register_range[-1]-register_range[0]+1), endian=self._endian)
        return_data = [(register, all_data[idx*8:idx*8+8]) for idx, register in enumerate(register_range)]
        return return_data
    return basic_data_getter_func

def _list_data_getter(register_name, n_bits):
    '''
    Function formula for getting a list-like register's data
    A list-like register contains repeated functionality every ``n_bits``
    over 1 or more register addresses.

    Returns a list of register_addr, bitarray pairs

    '''
    def list_data_getter_func(self):
        register_range = self.register_map[register_name]
        values = getattr(self, register_name)
        all_data = bitarray([])
        for value in values:
            all_data += bah.fromuint(value, n_bits, endian=self._endian)
        if len(all_data) < 8:
            all_data += bitarray([0]*(8-len(all_data)%8)) # pad up to full register
        return [(register_addr, all_data[idx*8:idx*8+8]) for idx, register_addr in enumerate(register_range)]
    return list_data_getter_func

def _compound_data_getter(registers):
    '''
    Function formula for getting a compound register's data
    A compound register has a one-to-many mapping of register address to
    register names.
    ``registers`` should be the register names contained within the
    register address (in order of lowest bit to highest)

    Returns a list of register, bitarray pairs

    '''
    def compound_data_getter_func(self):
        bits = bitarray([0]*8)
        for register_name in registers:
            start_bit, end_bit = self.bit_map[register_name]
            bits[start_bit%8:end_bit-start_bit+start_bit%8] = bah.fromuint(getattr(self, register_name), end_bit-start_bit, endian=self._endian)
        return [(self.register_map[registers[0]][0], bits)]
    return compound_data_getter_func

def _compound_list_data_getter(registers, n_bits):
    '''
    Function formula for getting a compound list register's data
    A compound list register has a one-to-many mapping of register
    addresses to register names, but with repeating functionality within
    each register name of size ``n_bits``. For example, you might store two
    1bit arrays of length 2 within the same register address.
    ``registers`` should be the register names contained within the
    register address (in order of lowest bit to highest)

    Returns a list of register, bitarray pairs

    '''
    def compound_list_data_getter_func(self):
        bits = bitarray([])
        for register_name in registers:
            values = getattr(self, register_name)
            for value in values:
                bits += bah.fromuint(value, n_bits, endian=self._endian)
        if len(bits)%8 != 0:
            bits += bitarray([0]*(8-len(bits)%8)) # pad up to full register
        return [(self.register_map[registers[0]][0], bits)]
    return compound_list_data_getter_func
# /Data getter function formulas

## Data setter function formulas
#
def _basic_data_setter(register_name):
    '''
    Function formula for setting a simple register's data
    Data can be specified as a register, bitarray pair or via a full-length
    bitarray

    '''
    def basic_data_setter_func(self, value):
        if self._is_register_value_pair(value):
            # regenerate complete list of bits and set
            set_register, set_bits = value
            data = getattr(self, register_name+'_data')
            all_bits = bitarray([])
            for register, bits in data:
                if register == set_register:
                    all_bits += set_bits
                else:
                    all_bits += bits
            start_bit, end_bit = self.bit_map[register_name]
            setattr(self, register_name+'_data', all_bits[start_bit%8:end_bit-start_bit+start_bit%8])
        else:
            # use all bits to set values
            bits = value
            value = bah.touint(bits, endian=self._endian)
            setattr(self, register_name, value)
    return basic_data_setter_func

def _list_data_setter(register_name, n_bits, min_value, max_value):
    '''
    Function formula for setting a list-like register's data
    A list-like register contains repeated functionality every ``n_bits``
    over 1 or more register addresses.
    Data can be specified as a register, bitarray pair or via a full-length
    bitarray

    '''
    def list_data_setter_func(self, values):
        if self._is_register_value_pair(values):
            # regenerate complete list of bits and set
            set_register, set_bits = values
            data = getattr(self, register_name+'_data')
            all_bits = bitarray([])
            for register, bits in data:
                if register == set_register:
                    all_bits += set_bits
                else:
                    all_bits += bits
            start_bit, end_bit = self.bit_map[register_name]
            setattr(self, register_name+'_data', all_bits[start_bit%8:end_bit-start_bit+start_bit%8])
        else:
            # use all bits to set values
            bits = values
            item_values = [bah.touint(bits[idx:idx+n_bits], endian=self._endian) for idx in range(0, len(bits), n_bits)]
            setattr(self, register_name, _Smart_List(item_values, min_value, max_value))
    return list_data_setter_func

def _compound_data_setter(registers, register_name):
    '''
    Function formula for setting a compound register's data
    A compound register has a one-to-many mapping of register address to
    register names.
    ``registers`` should be the register names contained within the
    register address (in order of lowest bit to highest)
    Data can be specified as a register, bitarray pair or via a full-length
    bitarray

    '''
    def compound_data_setter_func(self, value):
        if self._is_register_value_pair(value):
            set_register_addr, set_bits = value
            for register in registers:
                start_bit, end_bit = self.bit_map[register]
                setattr(self, register + '_data', set_bits[start_bit%8:end_bit-start_bit+start_bit%8])
        else:
            # use all bits to set values
            set_bits = value
            setattr(self, register_name, bah.touint(set_bits, endian=self._endian))
    return compound_data_setter_func

def _compound_list_data_setter(registers, register_name, n_bits, min_value, max_value):
    '''
    Function formula for setting a compound list register's data
    A compound list register has a one-to-many mapping of register address to
    register names with repeating functionality every ``n_bits``.
    ``registers`` should be the register names contained within the
    register address (in order of lowest bit to highest)
    Data can be specified as a register, bitarray pair or via a full-length
    bitarray

    '''
    def compound_list_data_setter_func(self, value):
        if self._is_register_value_pair(value):
            set_register_addr, set_bits = value
            for register in registers:
                start_bit, end_bit = self.bit_map[register]
                setattr(self, register + '_data', set_bits[start_bit%8:end_bit-start_bit+start_bit%8])
        else:
            set_bits = value
            values = [bah.touint(set_bits[idx:idx+n_bits], endian=self._endian) for idx in range(0,len(set_bits),n_bits)]
            setattr(self, register_name, _Smart_List(values, min_value, max_value))
    return compound_list_data_setter_func
# /Data setter function formulas

## Value validation function formulas
#
def _value_validator(value_types, min_value, max_value):
    '''
    Function formula for validating a register that contains a single value
    Accepts values of types ``value_types``, greater than  ``min_value``,
    and less than ``max_value``.

    '''
    def value_validator(func):
        @functools.wraps(func)
        def value_validated_func(self, value):
            if not isinstance(value, value_types):
                raise TypeError('value must be of type {}'.format(value_types))
            if value > max_value or value < min_value:
                raise ValueError('value must be between {} and {}'.format(min_value,max_value))
            return func(self, value)
        return value_validated_func
    return value_validator

def _list_validator(value_types, min_value, max_value, n_values):
    '''
    Function formula for validating a register that contains list-like data
    Accepts lists of length ``n_values`` with objects with types
    ``value_types``, greater than  ``min_value``, and less than ``max_value``.

    '''
    def list_validator(func):
        @functools.wraps(func)
        def list_validated_func(self, values):
            if not isinstance(values, (list, _Smart_List, tuple)):
                raise TypeError('argument must be a list or tuple')
            if len(values) != n_values:
                raise ValueError('length of list must be {}'.format(n_values))
            if any([not isinstance(value, value_types) for value in values]):
                raise TypeError('values must be of type {}'.format(value_types))
            if any([value > max_value or value < min_value for value in values]):
                raise ValueError('values must be between {} and {}'.format(min_value,max_value))
            return func(self, values)
        return list_validated_func
    return list_validator

def _data_validator(register_name):
    '''
    Function formula for validating setting a register via a register
    address, bitarray pair or direct bitarray

    '''
    def data_validator(func):
        @functools.wraps(func)
        def data_validated_func(self, value):
            if self._is_register_value_pair(value):
                register_addr, bits  = value
                if register_addr < self.register_map[register_name][0] \
                or register_addr > self.register_map[register_name][-1] \
                or len(bits) != 8:
                    raise ValueError('invalid register, value pair {}'
                        ' for register {}'.format(value,register_name))
                return func(self, value)
            elif isinstance(value, bitarray):
                bits = value
                start_bit, stop_bit = self.bit_map[register_name]
                if len(bits) != stop_bit - start_bit:
                    raise ValueError('invalid bitarray {}'
                        ' for register {}'.format(value,register_name))
                return func(self, value)
            else:
                raise TypeError('{} data must be assigned with '
                    'register, value pair or bitarray'.format(register_name))
        return data_validated_func
    return data_validator
# /Value validation function formulas

## Property function formulas
#
def _basic_property(name, registers, bits, types, min, max):
    docstring = '''
        simple value property

        registers: ``{}``

        valid types: ``{}``

        value range: ``{}`` to ``{}``
        '''.format(registers, types, min, max)
    return (
        property(
            _basic_getter(name),
            _value_validator(types,min,max)(
                _basic_setter(name)),
            doc=docstring),
        property(
            _basic_data_getter(name),
            _data_validator(name)(
                _basic_data_setter(name))))

def _list_property(name, registers, bits, types, min, max, length, n_bits):
    docstring = '''
        list-like property

        list length: ``{}``

        registers: ``{}``

        element valid types: ``{}``

        element value range: ``{}`` to ``{}``
        '''.format(length, registers, types, min, max)
    return (
        property(
            _basic_getter(name),
            _list_validator(types,min,max,length)(
                _list_setter(name,min,max)),
            doc=docstring),
        property(
            _list_data_getter(name,n_bits),
            _data_validator(name)(
                _list_data_setter(name,n_bits,min,max))))

def _compound_property(name, registers, bits, names, types, min, max):
    docstring = '''
        compound property

        registers: ``{}``

        bits: ``{}``

        valid types: ``{}``

        value range: ``{}`` to ``{}``

        shares a register with: ``{}``
        '''.format(registers, [bit%8 for bit in bits], types, min, max, [n for n in names if not n == name])
    return (
        property(
            _basic_getter(name),
            _value_validator(types,min,max)(
                _basic_setter(name)),
            doc=docstring
            ),
        property(
            _compound_data_getter(names),
            _data_validator(name)(
                _compound_data_setter(names, name))))

def _compound_list_property(name, registers, bits, names, types, min, max, length, n_bits):
    docstring = '''
        compound list-like property

        list length: ``{}``

        registers: ``{}``

        element valid types: ``{}``

        element value range: ``{}`` to ``{}``

        shares a register with: ``{}``
        '''.format(length, registers, types, min, max, [n for n in names if not n == name])
    return (
        property(
            _basic_getter(name),
            _list_validator(types,min,max,length)(
                _list_setter(name,min,max)),
            doc=docstring
            ),
        property(
            _compound_list_data_getter(names,n_bits),
            _data_validator(name)(
                _compound_list_data_setter(names,name,n_bits,min,max))))

def _generate_properties(cls, property_configuration, verbose=False):
    for _name, _prop_config in property_configuration.items():
        _prop_formula = _prop_config[0]
        _formula_args = _prop_config[1]
        _bit_range = _prop_config[2]
        if verbose:
            print('Generate {}.{} using bits {}\n\t{}({}) '.format(cls,_name, _bit_range, _prop_formula, _formula_args))

        # Add to class attributes
        cls.register_names += [_name]
        cls.bit_map[_name] = _bit_range
        cls.register_map[_name] = range(_bit_range[0]//8, max(_bit_range[1]//8,_bit_range[0]//8+1))

        # Create properties
        _prop, _prop_data = _prop_formula(_name, cls.register_map[_name],
            range(*cls.bit_map[_name]), *_formula_args)
        setattr(cls, _name, _prop)
        setattr(cls, _name+'_data', _prop_data)

    # Create a look up table from register address to names
    cls.register_map_inv = OrderedDict()
    for register_name, register_addr_range in cls.register_map.items():
        for register_addr in register_addr_range:
            try:
                cls.register_map_inv[register_addr] += [register_name]
            except KeyError:
                cls.register_map_inv[register_addr] = [register_name]


'''
Tests for basic logging functionality

'''
from __future__ import print_function
import pytest
import os
import numpy as np
from larpix.larpix import Packet_v1, Packet_v2, Controller, Chip, TimestampPacket
from larpix.io.fakeio import FakeIO
from larpix.logger.h5_logger import HDF5Logger

def test_enable(tmpdir):
    logger = HDF5Logger(directory=str(tmpdir))
    assert not logger.is_enabled()
    logger.record([Packet_v2()])
    assert len(logger._buffer['packets']) == 0
    logger.enable()
    assert logger.is_enabled()
    logger.record([Packet_v2()])
    assert len(logger._buffer['packets']) == 1

def test_disable(tmpdir):
    logger =HDF5Logger(directory=str(tmpdir), enabled=True)
    logger.disable()
    assert not logger.is_enabled()
    logger.record([Packet_v2()])
    assert len(logger._buffer['packets']) == 0

def test_flush(tmpdir):
    logger = HDF5Logger(directory=str(tmpdir), buffer_length=5,
            enabled=True)
    logger.record([Packet_v2()])
    assert len(logger._buffer['packets']) == 1
    logger.flush()
    assert len(logger._buffer['packets']) == 0
    logger.record([Packet_v2()]*5)
    assert len(logger._buffer['packets']) == 5
    logger.record([Packet_v2()])
    assert len(logger._buffer['packets']) == 0

def test_record(tmpdir):
    logger = HDF5Logger(directory=str(tmpdir), enabled=True)
    logger.record([Packet_v2()])
    assert len(logger._buffer['packets']) == 1
    logger.record([TimestampPacket(timestamp=123)])
    assert len(logger._buffer['packets']) == 2

@pytest.mark.filterwarnings("ignore:no IO object")
def test_controller_write_capture(tmpdir, chip):
    controller = Controller()
    controller.logger = HDF5Logger(directory=str(tmpdir), buffer_length=1)
    controller.logger.enable()
    controller.chips[chip.chip_key] = chip
    controller.write_configuration(chip.chip_key, 0)
    packet = chip.get_configuration_write_packets()[0]
    assert len(controller.logger._buffer['packets']) == 1

def test_controller_read_capture(tmpdir):
    controller = Controller()
    controller.io = FakeIO()
    controller.io.queue.append(([Packet_v2()], b'\x00\x00'))
    controller.logger = HDF5Logger(directory=str(tmpdir), buffer_length=1)
    controller.logger.enable()
    controller.run(0.1,'test')
    assert len(controller.logger._buffer['packets']) == 1



from larpix import Configuration_v2
from bitarray import bitarray
import larpix.bitarrayhelper as bah
from larpix import configs

import json
import random

def test_v2_conf_all_registers():
    default_filename = 'chip/default_v2.json'
    data = configs.load(default_filename)

    c = Configuration_v2()
    # check that all registers are in the configuration file
    assert set(data['register_values'].keys()) == set(c.register_names)
    for register in c.register_names:
        print('testing {}'.format(register))
        config_value = data['register_values'][register]
        stored_value = getattr(c,register)

        # test getters
        assert stored_value == config_value

        # test setters
        if register == 'chip_id':
            new_value = 12
        elif isinstance(stored_value, list):
            new_value = []
            for value in stored_value:
                if value == 1:
                    new_value += [0]
                elif value == 0:
                    new_value += [1]
                else:
                    new_value += [value-1]
        elif isinstance(stored_value, int):
            if stored_value == 1:
                new_value = 0
            elif stored_value == 0:
                new_value = 1
            else:
                new_value = stored_value-1
        setattr(c,register,new_value)
        assert getattr(c,register) == new_value

        # test data getters
        assert list(c.register_map[register]) == [reg for reg,bits in getattr(c,register+'_data')]

def test_conf():
    c = Configuration_v2()
    endian = 'little'

    # test simple register
    c.threshold_global = 255
    assert c.threshold_global == 255
    assert c.threshold_global_data == [(64, bah.fromuint(255,8, endian=endian))]

    c.threshold_global_data = (64, bah.fromuint(253,8, endian=endian))
    assert c.threshold_global == 253
    assert c.threshold_global_data == [(64, bah.fromuint(253,8, endian=endian))]

    c.threshold_global_data = bah.fromuint(252,8, endian=endian)
    assert c.threshold_global == 252
    assert c.threshold_global_data == [(64, bah.fromuint(252,8, endian=endian))]

    # test list register
    c.pixel_trim_dac = [0]*64
    assert c.pixel_trim_dac == [0]*64
    assert c.pixel_trim_dac_data == [(i, bah.fromuint(0,8, endian=endian)) for i in range(64)]

    c.pixel_trim_dac[1] = 1
    assert c.pixel_trim_dac[1] == 1
    assert c.pixel_trim_dac_data[1] == (1, bah.fromuint(1,8, endian=endian))
    assert c.pixel_trim_dac_data[0] == (0, bah.fromuint(0,8, endian=endian))

    c.pixel_trim_dac_data = (1, bah.fromuint(2,8, endian=endian))
    assert c.pixel_trim_dac[1] == 2
    assert c.pixel_trim_dac_data[1] == (1, bah.fromuint(2,8, endian=endian))
    assert c.pixel_trim_dac_data[0] == (0, bah.fromuint(0,8, endian=endian))

    bits = bitarray()
    for i in range(64):
        bits += bah.fromuint(31,8, endian=endian)
    c.pixel_trim_dac_data = bits
    assert c.pixel_trim_dac[1] == 31
    assert c.pixel_trim_dac[0] == 31
    assert c.pixel_trim_dac_data[1] == (1, bah.fromuint(31,8, endian=endian))
    assert c.pixel_trim_dac_data[0] == (0, bah.fromuint(31,8, endian=endian))

    # test compound register
    c.csa_gain = 1
    c.csa_bypass_enable = 1
    c.bypass_caps_en = 1
    reg_data = [(65, bitarray('11100000'))]
    assert c.csa_gain == 1
    assert c.csa_gain_data == reg_data
    assert c.csa_bypass_enable == 1
    assert c.csa_bypass_enable_data == reg_data
    assert c.bypass_caps_en == 1
    assert c.bypass_caps_en_data == reg_data

    c.csa_bypass_enable = 0
    reg_data = [(65, bitarray('10100000'))]
    assert c.csa_gain == 1
    assert c.csa_gain_data == reg_data
    assert c.csa_bypass_enable == 0
    assert c.csa_bypass_enable_data == reg_data
    assert c.bypass_caps_en == 1
    assert c.bypass_caps_en_data == reg_data

    reg_data = [(65, bitarray('01000000'))]
    c.csa_bypass_enable_data = reg_data[0]
    assert c.csa_gain == 0
    assert c.csa_gain_data == reg_data
    assert c.csa_bypass_enable == 1
    assert c.csa_bypass_enable_data == reg_data
    assert c.bypass_caps_en == 0
    assert c.bypass_caps_en_data == reg_data

    reg_data = [(65, bitarray('11000000'))]
    c.csa_gain_data = bitarray('1')
    assert c.csa_gain == 1
    assert c.csa_gain_data == reg_data
    assert c.csa_bypass_enable == 1
    assert c.csa_bypass_enable_data == reg_data
    assert c.bypass_caps_en == 0
    assert c.bypass_caps_en_data == reg_data

    # test list register (that covers <1 register)
    c.current_monitor_bank0 = [0]*4
    assert c.current_monitor_bank0 == [0]*4
    assert c.current_monitor_bank0_data == [(109, bah.fromuint(0,8, endian=endian))]

    c.current_monitor_bank0[1] = 1
    assert c.current_monitor_bank0[1] == 1
    assert c.current_monitor_bank0[0] == 0
    assert c.current_monitor_bank0_data == [(109, bitarray('01000000'))]

    c.current_monitor_bank0_data = (109, bitarray('00100000'))
    assert c.current_monitor_bank0[1] == 0
    assert c.current_monitor_bank0[2] == 1
    assert c.current_monitor_bank0_data == [(109, bitarray('00100000'))]

    bits = bitarray('0100')
    c.current_monitor_bank0_data = bits
    assert c.current_monitor_bank0[1] == 1
    assert c.current_monitor_bank0[0] == 0
    assert c.current_monitor_bank0_data == [(109, bitarray('01000000'))]

    # test compound list register (that covers <1 register)
    c.enable_miso_downstream = [0]*4
    c.enable_miso_differential = [0]*4
    assert c.enable_miso_differential == [0]*4
    assert c.enable_miso_differential_data == [(125, bah.fromuint(0,8, endian=endian))]

    c.enable_miso_differential[1] = 1
    assert c.enable_miso_differential[1] == 1
    assert c.enable_miso_differential[0] == 0
    assert c.enable_miso_differential_data == [(125, bitarray('00000100'))]

    c.enable_miso_differential_data = (125, bitarray('00000010'))
    assert c.enable_miso_differential[1] == 0
    assert c.enable_miso_differential[2] == 1
    assert c.enable_miso_differential_data == [(125, bitarray('00000010'))]

    bits = bitarray('0100')
    c.enable_miso_differential_data = bits
    assert c.enable_miso_differential[1] == 1
    assert c.enable_miso_differential[0] == 0
    assert c.enable_miso_differential_data == [(125, bitarray('00000100'))]

    # test long register
    c.periodic_trigger_cycles = 2**32-1
    assert c.periodic_trigger_cycles == 2**32-1
    assert c.periodic_trigger_cycles_data == [(i, bah.fromuint(255,8, endian=endian)) for i in range(166,170)]

    c.periodic_trigger_cycles_data = (166, bah.fromuint(254,8, endian=endian))
    assert c.periodic_trigger_cycles == 2**32-2
    assert c.periodic_trigger_cycles_data[0] == (166, bah.fromuint(254, 8, endian=endian))

    c.periodic_trigger_cycles_data = bah.fromuint(1,32, endian=endian)
    assert c.periodic_trigger_cycles == 1
    assert c.periodic_trigger_cycles_data[0] == (166, bah.fromuint(1,8, endian=endian))

def test_compare():
    c = Configuration_v2()
    other = Configuration_v2()
    assert c.compare(other) == {}

    c.pixel_trim_dac[10] = 25
    c.pixel_trim_dac[12] = 26
    assert c.compare(other) == {'pixel_trim_dac': [
        ({'index': 10, 'value': 25}, {'index': 10, 'value': 16}),
        ({'index': 12, 'value': 26}, {'index': 12, 'value': 16}),
        ]}

    c = Configuration_v2()
    c.threshold_global = 121
    assert c.compare(other) == {'threshold_global': (121, 255)}

def test_load_inheritance(tmpdir):
    filename = 'test.json'
    c = Configuration_v2()
    with open(str(tmpdir.join(filename)),'w') as of:
        config = {
            "_config_type": "chip",
            "_include": ["chip/default_v2.json"],
            "class": "Configuration_v2",
            "register_values": {
                "pixel_trim_dac": [0, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16],
                "threshold_global": 0
            }
        }
        json.dump(config,of)
    c.load(str(tmpdir.join(filename)))
    assert c.pixel_trim_dac[0] == 0
    assert c.threshold_global == 0
    assert c.csa_gain == 0

def test_get_nondefault_registers():
    c = Configuration_v2()
    c.pixel_trim_dac[10] = 25
    c.threshold_global = 121
    c.csa_gain = 1
    c.csa_enable[35] = 0
    assert c.get_nondefault_registers() == {
            'pixel_trim_dac': [({'index': 10, 'value': 25},
                {'index': 10, 'value': 16})],
            'threshold_global': (121, 255),
            'csa_gain': (1, 0),
            'csa_enable': [({'index': 35, 'value': 0}, {'index': 35,
                'value': 1})],
            }

def test_v2_conf_all_data():
    c = Configuration_v2()
    config_data = c.all_data()

    assert len(config_data) == c.num_registers, \
        f'config data length incorrect (should be {c.num_registers}, is {len(config_data)})'

    for reg_name in c.register_map.keys():
        reg_data = getattr(c, f'{reg_name}_data')
        for reg_addr, bits in reg_data:
            assert bits == config_data[reg_addr], \
                f'register {reg_addr} mismatch (should be {bits}, is {config_data[reg_addr]})'

def test_v2_conf_some_data():
    c = Configuration_v2()

    req_addrs = range(c.num_registers)
    addrs, bits = c.some_data(req_addrs)
    config_data = c.all_data()

    # check all addresses
    assert len(addrs) == len(req_addrs), \
        f'config data length incorrect (should be {len(req_addrs)}, is {len(addrs)})'
    assert set(addrs) == set(req_addrs), \
        f'config data values incorrect (should be {set(req_addrs)}, is {set(addrs)})'

    for addr, addr_bits in zip(addrs, bits):
        assert addr_bits == config_data[addr], \
            f'register {addr} mismatch (should be {config_data[addr]}, is {addr_bits})'

    # check partial addresses
    req_addrs = random.sample(range(c.num_registers), c.num_registers//2)
    addrs, bits = c.some_data(req_addrs)

    # check all addresses
    assert len(addrs) == len(req_addrs), \
        f'config data length incorrect (should be {len(req_addrs)}, is {len(addrs)})'
    assert set(addrs) == set(req_addrs), \
        f'config data values incorrect (should be {set(req_addrs)}, is {set(addrs)})'

    for addr, addr_bits in zip(addrs, bits):
        assert addr_bits == config_data[addr], \
            f'register {addr} mismatch (should be {config_data[addr]}, is {addr_bits})'



from bitarray import bitarray
import os
import errno
import functools
from collections import OrderedDict

from .. import bitarrayhelper as bah
from .. import configs

__all__ = [
    'BaseConfiguration',
    '_Smart_List',
]

class _Smart_List(list):
    '''
    A list type which checks its elements to be within given bounds.
    Used for Configuration attributes where there's a distinct value for
    each LArPix channel.

    '''

    def __init__(self, values, low, high):
        if not (type(values) == list or type(values) == _Smart_List):
            raise ValueError("_Smart_List is not list")
        if any([value > high or value < low for value in values]):
            raise ValueError("value out of bounds")
        list.__init__(self, values)
        self.low = low
        self.high = high

    def __setitem__(self, key, value):
        if isinstance(key, int):
            if value > self.high or value < self.low:
                raise ValueError("value out of bounds")
            list.__setitem__(self, key, value)
        else:
            for num in value:
                if num > self.high or num < self.low:
                    raise ValueError("value out of bounds")
            list.__setitem__(self, key, value)

    def __setslice__(self, i, j, value):
        '''
        Only used in Python 2, where __setslice__ is deprecated but
        contaminates the namespace of this subclass.

        '''
        self.__setitem__(slice(i, j, None), value)

def _nice_json(d, depth=1):
    '''
    Helper method for printing dicts in a dense but readable format for json

    '''
    l = []
    for key,value in d.items():
        if isinstance(value,str):
            l += ['\"{}\": \"{}\"'.format(key,value)]
        elif isinstance(value,dict):
            l += ['\"{}\": {}'.format(key, _nice_json(value, depth+1))]
        elif isinstance(value,bool):
            if value:
                l += ['\"{}\": true'.format(key)]
            else:
                l += ['\"{}\": false'.format(key)]
        elif value is None:
            l += ['\"{}\": null'.format(key)]
        else:
            l += ['\"{}\": {}'.format(key,value)]
    return '{\n'+depth*'    ' + (',\n'+depth*'    ').join(l) + '\n' +(depth-1)*'    '+'}'


class BaseConfiguration(object):
    '''
    Base class for larpix configuration objects

    '''

    def __init__(self):
        self.load(self.default_configuration_file)

    def __setattr__(self, name, value):
        '''
        Default setattr behavior occurs if name is in ``register_names``, is "private"
        or is a known attribute
        Otherwise raises an attribute error

        '''
        # print('SET', name, value)
        if not (name in self.register_names or name[0] == '_' or hasattr(self, name)):
            raise AttributeError('%s is not a known register' % name)
        return super(BaseConfiguration, self).__setattr__(name, value)

    def __eq__(self, other):
        '''
        Returns true if all fields match
        '''
        return all([getattr(self, register_name) == getattr(other, register_name)
                    for register_name in self.register_names])

    def __str__(self):
        '''
        Converts configuration to a nicely formatted json string

        '''
        d = self.to_dict()
        return _nice_json(d)

    def compare(self, other):
        '''
        Returns a dict containing pairs of each differently valued register
        Pair order is (self, other)
        '''
        d = {}
        for register_name in self.register_names:
            if getattr(self, register_name) != getattr(other, register_name):
                d[register_name] = (getattr(self, register_name), getattr(other,
                    register_name))
        # Attempt to simplify some of the long values (array values)
        for (name, (self_value, config_value)) in d.items():
            if isinstance(self_value,(list,_Smart_List)):
                different_values = []
                for ch, (val, config_val) in enumerate(zip(self_value, config_value)):
                    if val != config_val:
                        different_values.append(({'index': ch, 'value': val},
                                                 {'index': ch, 'value': config_val}))
                if len(different_values) < 5:
                    d[name] = different_values
                else:
                    pass
        return d

    def get_nondefault_registers(self):
        '''
        Return a dict of all registers that are not set to the default
        configuration (i.e. of the ASIC on power-up). The keys are the
        register name where there's a difference, and the values are
        tuples of (current, default) configuration values.

        '''
        return self.compare(self.__class__())

    def to_dict(self):
        '''
        Export the configuration register names and values into a dict.

        '''
        d = {}
        for register_name in self.register_names:
            d[register_name] = getattr(self, register_name)
        return d

    def from_dict(self, d):
        '''
        Use a dict of ``{register_name, value}`` to update the current
        configuration. Not all registers must be in the dict - only
        those present will be updated.

        '''
        for register_name in self.register_names:
            if register_name in d:
                setattr(self, register_name, d[register_name])

    def write(self, filename, force=False, append=False):
        '''
        Save the configuration to a JSON file.

        '''
        if os.path.isfile(filename):
            if not force:
                raise IOError(errno.EEXIST,
                              'File %s exists. Use force=True to overwrite'
                              % filename)
        d = {
            '_config_type': 'chip',
            'class': self.__class__.__name__,
            'register_values': self.to_dict()
        }
        with open(filename, 'w+') as outfile:
            outfile.write(_nice_json(d))
        return 0

    def load(self, filename):
        '''
        Load a JSON file and use the contents to update the current
        configuration.

        '''
        data = configs.load(filename, 'chip')
        if data['class'] != self.__class__.__name__:
            raise RuntimeError('Configuration is not of class {}'.format(data['class']))
        self.from_dict(data['register_values'])



'''
The setup.py file for larpix-control.

'''

from setuptools import setup, find_packages
from codecs import open
import os

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(os.path.join(here, 'VERSION')) as f:
    version = f.read()

setup(
    name='larpix-control',
    version=version,
    description='Control the LArPix chip',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Other/Proprietary License',
    url='https://github.com/larpix/larpix-control',
    author='Peter Madigan and Sam Kohn',
    author_email='pmadigan@lbl.gov',
    classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3'
    ],
    keywords='dune physics',
    packages=find_packages(),
    install_requires=[
        'pyserial >=3.4',
        'pytest >=6.2',
        'bitarray >=0.8',
        'pyzmq >= 18.0',
        'sphinx_rtd_theme >= 0.5',
        'numpy >= 1.16',
        'h5py >= 3.1',
        'bidict >= 0.18.0',
        'networkx >= 2.2'
    ],
    scripts=[
        'scripts/gen_controller_config.py',
        'scripts/gen_hydra_simple.py',
        'scripts/convert_rawhdf5_to_hdf5.py',
        'scripts/packet_hdf5_tool.py',
        'scripts/raw_hdf5_tool.py'
    ],
    package_data={
        'larpix.configs': ['*/*.json'],
    },
)



'''
Tests for larpix.io.zmq_io module

'''
from __future__ import print_function
import pytest
import os
import json
from larpix.larpix import Packet, Packet_v1, Key
from larpix.io.zmq_io import ZMQ_IO
from larpix.format.message_format import dataserver_message_encode

@pytest.fixture
def io_config(tmpdir):
    filename = str(tmpdir.join('test_conf.json'))
    config_dict = {
            "_config_type": "io",
            "io_class": "ZMQ_IO",
            "io_group": [
                [1, "192.0.2.1"]
            ]
        }
    with open(filename,'w') as of:
        json.dump(config_dict, of)
    return filename

@pytest.fixture
def zmq_io_obj(io_config):
    return ZMQ_IO(io_config)

def test_encode(zmq_io_obj):
    io_chain = 1
    test_packet = Packet(b'\x01'*Packet.num_bytes)
    test_packet.chip_key = Key(1, io_chain, test_packet.chip_id)
    test_bytes = b'0x0101010101010101 1'
    if isinstance(test_packet, Packet_v1):
        test_bytes = b'0x0001010101010101 1'
    expected = [test_bytes]
    assert zmq_io_obj.encode([test_packet]) == expected

def test_decode(zmq_io_obj):
    io_chain = 1
    test_packet = Packet(b'\x01'*Packet.num_bytes)
    test_packet.chip_key = Key(1, io_chain, test_packet.chip_id)
    test_bytes = dataserver_message_encode([test_packet])
    expected = [test_packet]
    assert zmq_io_obj.decode(test_bytes) == expected



from collections import defaultdict

from larpix import Key

class SyncPacket(object):
    '''
    A packet-like object which contains "sync" data (e.g. associated with
    sync words in the PACMAN ZMQ message format).

    It implements the core methods used by Packet, so it works seamlessly in
    lists of packets and in a PacketCollection.

    :param sync_type: optional, an 8-bit type indication for the sync type
    
    :param clk_source: optional, a 1-bit indicator used in clock switch packets

    :param timestamp: optional, a 32-bit timestamp

    :param io_group: optional, an 8-bit io_group id

    '''
    packet_type = 6
    
    pretty_sync_type = defaultdict(lambda:'OTHER')
    pretty_sync_type[b'S'] = 'SYNC'
    pretty_sync_type[b'H'] = 'HEARTBEAT'
    pretty_sync_type[b'C'] = 'CLOCK SWITCH'
    
    def __init__(self, sync_type=None, clk_source=None, timestamp=None, io_group=None):
        self.sync_type = sync_type
        self.clk_source = clk_source
        self.timestamp = timestamp
        self.io_group = io_group

    def __eq__(self, other):
        if self.sync_type != other.sync_type: return False
        if self.clk_source != other.clk_source: return False
        if self.timestamp != other.timestamp: return False
        if self.io_group != other.io_group: return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        strings = ['Sync']
        if self.io_group is not None:
            strings.append('IO group: {}'.format(self.io_group))        
        if self.sync_type is not None:
            strings.append('Type: {}'.format(self.pretty_sync_type[self.sync_type]))
        if self.timestamp is not None:
            strings.append('Timestamp: {}'.format(self.timestamp))
        if self.clk_source is not None:
            strings.append('Clk source: {}'.format(self.clk_source))
        return '[ '+' | '.join(strings)+' ]'

    def __repr__(self):
        strings = list()
        if self.io_group is not None:
            strings.append('io_group={}'.format(self.io_group))        
        if self.sync_type is not None:
            strings.append('sync_type={}'.format(self.sync_type))
        if self.timestamp is not None:
            strings.append('timestamp={}'.format(self.timestamp))
        if self.clk_source is not None:
            strings.append('clk_source={}'.format(self.clk_source))
        return 'SyncPacket('+', '.join(strings)+')'

    def export(self):
        d = dict()
        d['io_group'] = self.io_group
        d['sync_type'] = self.sync_type
        d['timestamp'] = self.timestamp
        d['clk_source'] = self.clk_source
        d['type'] = self.packet_type
        return d

    def from_dict(self, d):
        if 'io_group' in d:
            self.io_group = d['io_group']
        if 'sync_type' in d:
            self.sync_type = d['sync_type']
        if 'timestamp' in d:
            self.timestamp = d['timestamp']
        if 'clk_source' in d:
            self.clk_source = d['clk_source']

    @property
    def chip_key(self):
        if self.io_group:
            return Key(self.io_group,0,0)
        return None

    @chip_key.setter
    def chip_key(self, val):
        key = Key(val)
        self.io_group = key.io_group



from .packet_v1 import *
from .packet_v2 import *
from .timestamp_packet import *
from .message_packet import *
from .trigger_packet import *
from .sync_packet import *
Packet = Packet_v2

from .packet_collection import *



from .key import Key
from .configuration import Configuration_v1, Configuration_v2, Configuration_v2b, Configuration_Lightpix_v1
from .packet import Packet_v1, Packet_v2

class Chip(object):
    '''
    Represents one LArPix chip and helps with configuration and packet
    generation.

    '''

    def __init__(self, chip_key, version=2):
        '''
        Create a new Chip object with the given ``chip_key``. See the ``Key``
        class for the key specification. Key can be specified by a valid keystring
        or a ``Key`` object.

        '''
        self.asic_version = version
        if self.asic_version == 1:
            self.config = Configuration_v1()
        elif self.asic_version == 2:
            self.config = Configuration_v2()
        elif self.asic_version == 'lightpix-v1.0':
            self.config = Configuration_Lightpix_v1()
        elif self.asic_version == '2b':
            self.config = Configuration_v2b()
        else:
            raise RuntimeError('chip asic version is invalid')
        chip_key = Key(chip_key)
        self.io_group = chip_key.io_group
        self.io_channel = chip_key.io_channel
        self.chip_id = chip_key.chip_id
        self.reads = []
        self.new_reads_index = 0

    def __str__(self):
        return 'Chip (key: {}, version: {})'.format(str(self.chip_key), self.asic_version)

    def __repr__(self):
        return 'Chip(chip_key={}, version={})'.format(str(self.chip_key), self.asic_version)

    @property
    def chip_key(self):
        return Key(self.io_group, self.io_channel, self.chip_id)

    @chip_key.setter
    def chip_key(self, val):
        val = Key(val)
        self.io_group = val.io_group
        self.io_channel = val.io_channel
        self.chip_id = val.chip_id

    def is_chip_id_set(self):
        '''
        Check if chip id (as specified by unique key) matches the chip id stored
        in the configuration.
        Only valid for v2 asics, if v1 asic, will always return ``True``
        Note: Even if this function returns True, it's possible the chip ID has not been sent from the larpix-control software onto the ASIC hardware

        '''
        if self.asic_version == 1:
            return True
        elif self.asic_version in (2, 'lightpix-v1.0'):
            return self.config.chip_id == self.chip_id

    def get_configuration_packets(self, packet_type, registers=None):
        '''
        Return a list of Packet objects to read or write (depending on
        ``packet_type``) the specified configuration registers (or all
        registers by default).

        '''
        conf = self.config
        if registers is None:
            registers = range(conf.num_registers)
        packets = []
        if self.asic_version == 1:
            packet_register_data = enumerate(conf.all_data())
        elif self.asic_version in (2, 'lightpix-v1.0', '2b'):
            packet_register_data = zip(*conf.some_data(registers))
        for i, data in packet_register_data:
            if i not in registers:
                continue
            if self.asic_version == 1:
                packet = Packet_v1()
            else:
                packet = Packet_v2()
            packet.packet_type = packet_type
            packet.chip_id = self.chip_id
            packet.chip_key = self.chip_key
            packet.register_address = i
            if packet_type == packet.CONFIG_WRITE_PACKET:
                packet.register_data = data
            elif packet_type == packet.CONFIG_READ_PACKET:
                packet.register_data = 0
            else:
                raise ValueError('incorrect packet_type for configuration packets')
            packet.assign_parity()
            packets.append(packet)
        return packets

    def get_configuration_write_packets(self, registers=None):
        '''
        Return a list of Packet objects to write corresponding to the specified
        configuration registers (all by default)

        '''
        if self.asic_version == 1:
            return self.get_configuration_packets(Packet_v1.CONFIG_WRITE_PACKET, registers)
        return self.get_configuration_packets(Packet_v2.CONFIG_WRITE_PACKET, registers)

    def get_configuration_read_packets(self, registers=None):
        '''
        Return a list of Packet objects to write corresponding to the specified
        configuration registers (all by default)

        '''
        if self.asic_version == 1:
            return self.get_configuration_packets(Packet_v1.CONFIG_READ_PACKET, registers)
        return self.get_configuration_packets(Packet_v2.CONFIG_READ_PACKET, registers)

    def sync_configuration(self, index=-1):
        '''
        Adjust self.config to match whatever config read packets are in
        self.reads[index].

        Defaults to the most recently read PacketCollection. Later
        packets in the list will overwrite earlier packets. The
        ``index`` parameter could be a slice.

        '''
        updates = {}
        if isinstance(index, slice):
            for collection in self.reads[index]:
                for packet in collection:
                    if packet.packet_type == packet.CONFIG_READ_PACKET:
                        updates[packet.register_address] = packet.register_data
        else:
            for packet in self.reads[index]:
                if packet.packet_type == packet.CONFIG_READ_PACKET:
                    updates[packet.register_address] = packet.register_data
        if self.asic_version == 1:
            self.config.from_dict_registers(updates)
        else:
            self.config.from_dict_registers(updates, endian=Packet_v2.endian)

    def export_reads(self, only_new_reads=True):
        '''
        Return a dict of the packets this Chip has received.

        If ``only_new_reads`` is ``True`` (default), then only the
        packets since the last time this method was called will be in
        the dict. Otherwise, all of the packets stored in ``self.reads``
        will be in the dict.

        '''
        data = {}
        data['chip_key'] = self.chip_key
        data['chip_id'] = self.chip_id
        if only_new_reads:
            packets = self.reads[self.new_reads_index:]
        else:
            packets = self.reads
        data['packets'] = list(map(lambda x:x.export(), packets))
        self.new_reads_index = len(self.reads)
        return data



from larpix.larpix import Chip, Packet, TimestampPacket
from larpix.format.message_format import dataserver_message_decode, dataserver_message_encode

def test_message_format_test_packets(chip):
    expected_packets = chip.get_configuration_packets(Packet.CONFIG_READ_PACKET)
    expected_messages = [b'\x01\x00'+b'D'+b'\x02'+b'\x00'*4 + packet.bytes() for packet in expected_packets]
    ts_packet = TimestampPacket(timestamp=123456789)
    expected_packets += [ts_packet]
    expected_messages += [b'\x01\x00'+b'T'+b'\x00'*5 + ts_packet.bytes() + b'\x00']
    print(expected_packets[-1])
    print(expected_messages[-1])
    print(dataserver_message_decode(expected_messages)[-1])
    print(expected_messages[-1])
    print(dataserver_message_encode(expected_packets)[-1])
    assert expected_messages == dataserver_message_encode(expected_packets)
    assert expected_packets == dataserver_message_decode(expected_messages)
